#!/usr/bin/env bash
#===============================================================
# MUSASHI bash script
#
# 顧客の購買履歴データを作成する
#
# Version :  2002/05/22 初期作成 by Y.Hamuro
#
# 入力ファイル : $outPath/syo.xt
# 出力ファイル : $outPath/dat.xt
#              : $outPath/cust.xt
#
# メイン処理の"store A 300"の行を別の店名で複写するとより多くの
# 店のデータを作成できる。また顧客人数を増やすことによって、
# 店あたりのデータ件数を増やすことができる。
#
# 出力ファイルのイメージ
#  店 日付 時間 レシート 顧客 商品 大分類 中分類 小分類 細分類 メーカー ブランド 仕入単価 単価 数量 金額 仕入金額 粗利金額
#  A 20020109 193634 1000002 A00001 0000234 1 13 1304 130403 0498 049803 182 239 1 239 182 57
#  A 20020109 193634 1000002 A00001 0000338 1 11 1119 111909 0270 027002 168 214 1 214 168 46
#  A 20020109 193634 1000002 A00001 0000352 1 11 1102 110209 1447 144703 323 396 1 396 323 73
#  A 20020109 193634 1000002 A00001 0000479 1 11 1106 110605 1270 127005 310 473 1 473 310 163
#  A 20020109 193634 1000002 A00001 0000573 1 14 1403 140321 0391 039101 423 579 4 2316 1692 624
#  A 20020113 170104 1000003 A00002 0000278 1 11 1104 110407 1837 183704 254 277 2 554 508 46
#  A 20020113 170104 1000003 A00002 0000389 1 11 1106 110605 1372 137204 278 401 1 401 278 123
#  A 20020227 075716 1000004 A00007 0000158 1 11 1103 110303 0629 062901 309 401 5 2005 1545 460
#  A 20020227 075716 1000004 A00007 0000168 1 11 1115 111509 1836 183604 310 390 1 390 310 80
#  A 20020227 075716 1000004 A00007 0000193 1 11 1121 112107 1352 135205 303 398 1 398 303 95

#===============================================================
inPath="../indat"
outPath="../outdat"

. ./debugEnv.sh

#---- タイトル
title="顧客購買履歴データ"

#---- コメント
comment="人工データ"

#---- 出力ファイル名
#outFile="dat/dat.xt"

#---- 乱数の種
RandSeed=10

#---------------------------------------------------------------
# コマンド
#---------------------------------------------------------------

#===============================================================================
#各顧客ごとに購入履歴を作成する関数
##### 入力パラメータ
# $1:店
# $2:顧客
# $3:来店間隔
# $4:平均購入数量
# $5:購入数量標準偏差
# $6:来店開始日付
# $7:来店終了日付
#
##### 出力ファイル(顧客別販売明細,"tmp/custXXXXXX")
# 1:店
# 2:日付
# 3:時間
# 4:顧客(店＋５桁の数字,"A00001")
# 5:商品(7桁の数字,"0000645")
#-------------------------------------------------------------------------------
function cust {
  #2002/01/01から１年間の日付(365行)を作成
  ${dbg}xtmkdata -l 365 -c 'dayadd("20020101",line()-1)' -a 日付 |

  #来店間隔に基づき、一様乱数で顧客の来店日を選択する
  #(0〜来店間隔の乱数を生成し、0であればその行を選択)
  ${dbg}xtsel -c "rand(0,$3,$RandSeed)<1" |

  #選択された日付が来店開始日付以降,来店終了日付以前である日付のみ選択
  ${dbg}xtsel -c '$日付>='$6 |
  ${dbg}xtsel -c '$日付<='$7 |

  #11時と16時が来店のピークとなるような来店時間を求める
  #11時もしくは16時に,平均0,標準偏差2時間(7200)に従う正規分布乱数を足した時間
  ${dbg}xtcal -c "if(rand(0,2,$RandSeed)==0,timeadd("110000",nrand(0,7200,$RandSeed)),timeadd("160000",nrand(0,7200,$RandSeed)))" -a 時間 |

  #各日付における購入数量を求める
  #(パラメータで与えられた平均購入数量、購入数量標準偏差に従う正規分布乱数にて)
  ${dbg}xtcal -c "up(nrand($4,$5,$RandSeed),1)" -a 購入数量 |

  #購入数量が０以下であれば省く
  ${dbg}xtsel -c "$購入数量>0" |

  #------------------------------------------------------------
  #ここまでである顧客の日付,時間,購入数量が求まったことになる。
  # 20020401 114513 5
  # 20020515 144521 3
  #     :      :    :
  #------------------------------------------------------------

  #各行の数量分の行を作成する
  ${dbg}xtduprec -f 購入数量 |

  #購入商品を正規分布乱数によって決定する
  ${dbg}xtcal -c "up(nrand($syoAvg,$syoStd,$RandSeed),1)" -a 商品乱数 |
  ${dbg}xtsel -c '$商品乱数>0 && $商品乱数<='$syoCnt |
  ${dbg}xtcal -c 'fix($商品乱数,7,"R","0")' -a 商品 |

  #顧客と店の項目を付ける
  ${dbg}xtsetchr -v $2 -a 顧客 |
  ${dbg}xtsetchr -v $1 -a 店 |

  #tmpディレクトリに顧客の名前で保存
  ${dbg}xtcut -f 店,日付,時間,顧客,商品 >tmp/cust$2.xt
}

#===============================================================================
#ある店の顧客リストを作成し、そのリストに基づいて顧客履歴作成関数(cust)を起動し
#その後に、全顧客の履歴をまとめ、店ファイルを作成する。
#-------------------------------------------------------------------------------
##### 入力パラメータ
#$1:店
#$2:人数

##### 顧客履歴作成関数(cust)起動ファイル
# 0:"scp"
# 1:店
# 2:顧客(店＋５桁の数字,"A00001")
# 3:来店間隔(平均値30日、標準偏差10日の正規分布に従った乱数)
# 4:平均購入数量(平均値5個、標準偏差2個の正規分布に従った乱数)
# 5:購入数量標準偏差(3個に固定)
# 6:来店開始日付(2001/07/01を中心として平均値0標準偏差200の乱数を足した日付)
#   この日付以前の来店がないという意味
# 7:来店終了日付(2003/07/01を中心として平均値0標準偏差200の乱数を足した日付)
#   この日付以降の来店がないという意味

##### 出力ファイル(店別販売明細,"tmp/storeX")
# 1:店
# 2:日付
# 3:時間
# 4:顧客(店＋５桁の数字,"A00001")
# 5:商品(7桁の数字,"0000645")
# 6:レシート番号(7桁の数字,"1000012")
#===============================================================================
function store {
  #顧客別ワークファイルの削除
  rm -f tmp/cust*.xt

  #５桁の顧客番号
  ${dbg}xtmkdata -l $2 -c 'fix(line(),5,"R","0")' -a 顧客番号 |

  #店							cust関数の第１パラメータ
  ${dbg}xtsetchr -v $1 -a 店 |

  #顧客(店+顧客番号)					cust関数の第２パラメータ
  ${dbg}xtcal -c 'cat($店,$顧客番号,"")' -a 顧客 |

  #来店間隔(日数)					cust関数の第３パラメータ
  ${dbg}xtcal    -c "up(nrand(30,10,$RandSeed),1)" -a 来店間隔 |

  #平均購入数量						cust関数の第４パラメータ
  ${dbg}xtcal    -c "up(nrand(5,2,$RandSeed),1)" -a 平均購入数量 |

  #購入数量標準偏差					cust関数の第５パラメータ
  ${dbg}xtsetchr -v 3 -a 購入数量標準偏差 |

  #来店開始日付						cust関数の第６パラメータ
  ${dbg}xtcal -c "dayadd("20010701",abs(nrand(0,200,$RandSeed)))" -a 来店開始日付 |

  #来店終了日付						cust関数の第７パラメータ
  ${dbg}xtcal -c "dayadd("20030701",-1*abs(nrand(0,200,$RandSeed)))" -a 来店終了日付 |

  #顧客履歴作成関数の名前をセット
  ${dbg}xtsetchr -v cust -a 関数名 |

  #以下の条件を持つ顧客のみ選択する
  ${dbg}xtsel -c '$来店間隔>2' |
  ${dbg}xtsel -c '$平均購入数量>0' |
  ${dbg}xtsel -c '$来店開始日付<$来店終了日付'  |

  #必要な項目をパラメータの順序に並べかえて選択
  ${dbg}xtcut -f 関数名,店,顧客,来店間隔,平均購入数量,購入数量標準偏差,来店開始日付,来店終了日付 |

  #shスクリプト(テキスト)に変換する
  ${dbg}xt2txt >xxscp

  #顧客履歴作成関数の起動(スクリプトの実行)
  source xxscp

  #cust関数で作成された顧客履歴を一つにまとめる
  ${dbg}xtcat -i 'tmp/cust*.xt' |

  #レシート番号を付ける
  ${dbg}xtcal -k 日付,顧客 -c 'keyNo()+1000000' -a レシート >tmp/store$1.xt

  #顧客別ワークファイルの削除
  rm -f tmp/cust*.xt
}

#===============================================================================
#顧客マスターの作成
#-------------------------------------------------------------------------------
##### 入力パラメータ
#$1:全購買履歴ファイル名
#$2:顧客マスター名

##### 出力ファイル(顧客マスター,"$outPath/cust.xt")
# 1:顧客
# 2:生年月日
# 3:性別("男","女")
#===============================================================================
function custMaster {
  ${dbg}xtcut -f 顧客 -i $1 |
  ${dbg}xtuniq -k 顧客 |

  #生年月日
  ${dbg}xtcal -c "dayadd(cat(1960+up(nrand(0,10,$RandSeed),1),"0101",""),rand(0,365,$RandSeed))" -a 生年月日 |

  #性別 (男：女＝１：５の割合)
  ${dbg}xtcal -c "if(rand(0,5,$RandSeed)<1,男,女)" -a 性別 |
  ${dbg}xtheader -l "顧客マスター" -o $2
}

#===============================================================================
#単価マスターの作成
#店別商品別日付別の単価計算
#-------------------------------------------------------------------------------
##### 入力パラメータ
#$1:全購買履歴ファイル名
#$2:単価マスター名

##### 出力ファイル(顧客マスター,"$outPath/cust.xt")
# 1:店
# 2:商品
# 3:日付
# 4:単価
#===============================================================================
function price {
  #店別商品別日付別の仕入単価表の作成(同じ商品は同じ仕入単価)
  ${dbg}xtjoin -k 商品 -m $outPath/syo.xt -f 仕入単価 -i $1 |
  ${dbg}xtcut  -f 店,商品,日付,仕入単価 |
  ${dbg}xtsort -k 店,商品,日付 |
  ${dbg}xtuniq -k 店,商品,日付 |
  ${dbg}xtsort -k 店,商品,日付 |

  #単価計算
  #ある商品について、
  #最初の日付における単価は仕入単価＊nrand(1.3)で計算される。
  #次の行(日付)からの単価は、前の行(日付)の単価から計算する。
  #90%の確率で前の行と同じ単価を用いる。
  #10%の確率で単価は変動し、前の行の単価に±10%を加える。
  ${dbg}xtcal  -k 店,商品 -c '
    if(keyLine()==1,
      up($仕入単価*(1+nrand(0.3,0.1,'$RandSeed')),1),
      if(rand(0,10,'$RandSeed')<9,
        prvResult(),
        prvResult()+down(nrand(0,$仕入単価*0.1,'$RandSeed'),1)
      )
    )' -a 単価 -o $2
}

#===============================================================================
# メイン処理
#===============================================================================
#--------------------------------------------
#ワークファイルを入れるディレクトリ作成
#--------------------------------------------
mkdir tmp 2>/dev/null

#--------------------------------------------
#商品コード決定のために利用する変数
#--------------------------------------------
#syoCnt : 商品数
#syoAvg : 商品コードの平均
#syoStd : 商品コードの標準偏差
syoCnt=`xt2txt -i $outPath/syo.xt | wc -l`
syoCnt=$((syoCnt+0))
syoAvg=$((syoCnt/2))
syoStd=$((syoCnt/6))

#--------------------------------------------
#店別明細の作成
#--------------------------------------------
#関数名,店,顧客人数
#store A 280
#store B 300
#store C 250
#store D 260
store A 10
store B 10
store C 10
store D 10

#店別ファイルを一つにまとめる
${dbg}xtcat -i 'tmp/store*.xt' |
${dbg}xtcut -f 店,日付,時間,レシート,顧客,商品 >xxbase

rm -rf tmp

#--------------------------------------------
#顧客マスターの作成
#--------------------------------------------
custMaster xxbase $outPath/cust.xt

#--------------------------------------------
#単価マスターの作成
#--------------------------------------------
price      xxbase xxprice

#--------------------------------------------
#join+数量金額の計算
#--------------------------------------------
#商品マスター項目のjoin
${dbg}xtjoin -k 商品 -m $outPath/syo.xt -f 大分類,中分類,小分類,細分類,メーカー,ブランド,仕入単価 -i xxbase |

#単価のjoin
${dbg}xtjoin -k 店,商品,日付 -m xxprice -f 単価 |

#数量計算
#80%の確率で１個、else 50%の確率で２個 else nrand(4,1)
${dbg}xtcal -c "if(rand(0,10,$RandSeed)<8,1,if(rand(0,10,$RandSeed)<5,2,up(nrand(4,1,$RandSeed),1)))" -a 数量 |

#金額、仕入金額、粗利金額の計算
${dbg}xtcal -c '$単価*$数量' -a 金額 |
${dbg}xtcal -c '$仕入単価*$数量' -a 仕入金額 |
${dbg}xtcal -c '($単価-$仕入単価)*$数量' -a 粗利金額 |

#並べかえてヘッダーを変更
${dbg}xtsort -k 店,日付,時間,レシート |
${dbg}xtheader -l "$title" -c "$comment" -o $outPath/dat.xt

#ワークファイルを削除する
rm -f xx*
#===============================================================
